import rclpy
import os
from modern_robotics import FKinBody
from ament_index_python.packages import get_package_share_directory
from mr_urdf_loader import loadURDF
from rclpy.node import Node
from math import sin, cos, pi
import numpy as np
from std_msgs.msg import Float64
from nav_msgs.msg import Path
from geometry_msgs.msg import PoseStamped
from geometry_msgs.msg import Pose
from visualization_msgs.msg import Marker
import transformations
from scipy.spatial.transform import Rotation
import time


class Forward_Kinematics_Node_2Dof(Node):

    def __init__(self):
        super().__init__('forward_kinematics_node_2dof')
        # Create a subscriber, initialising the transformation
        self.subscription = self.create_subscription(
            Float64,
            '/dof2/reference_angles',
            self.listener_callback,
            10)
        
        urdf_file_path = os.path.join(
            get_package_share_directory('iiwa_description'), 'urdf', 'iiwa14.urdf')
        self.robot_model = loadURDF(urdf_file_path)
        self.M = self.robot_model["M"]
        self.Slist = self.robot_model["Slist"]
        self.Mlist = self.robot_model["Mlist"]
        self.Glist = self.robot_model["Glist"]
        self.Blist = self.robot_model["Blist"]

        # Create two publishers for the RViz simulation - one for the whole path and one for the marker
        self.path_publisher_ = self.create_publisher(Path, '/simple_robot/visualization_path', 10)
        self.point_publisher_ = self.create_publisher(Marker, '/simple_robot/visualization_marker', 10)
        # create a publisher for the plotjuggler data
        self.angle_publisher_ = self.create_publisher(Float64, '/simple_robot/angle_2dof/plotjuggler', 10)
        self.height_publisher_ = self.create_publisher(Float64, '/simple_robot/height_2dof/plotjuggler', 10)
        # create a publisher for the pose
        self.pose_publisher_simple_robot_ = self.create_publisher(PoseStamped, '/simple_robot/end_effector_pose', 10)
        self.pose_publisher_kuka_robot_ = self.create_publisher(Pose, '/kuka_iiwa/target_pose', 10)
        #RViz
        self.path = Path()
        self.path.header.frame_id = "map"
        self.path.header.stamp = self.get_clock().now().to_msg()

        self.index = 0

    def listener_callback(self, msg):
        # Get the message
        self.phi = msg.data

        if self.phi == 0:
            self.index += 1
        
        # call the transformation function
        self.transformation()
        
    def transformation(self):
        # Define the desired rotation 
        theta = pi / 1.8
        # Split theta
        theta_fraction = self.phi * theta

        self.publish_angle(theta_fraction)

        # Definte the desired height
        h = 0.3
        #Split h
        h_fraction = self.phi * h

        self.publish_height(h_fraction)

        # Define the matrix
        intermediate_rotation = np.array([
            [cos(theta_fraction), -sin(theta_fraction), 0, 0],
            [sin(theta_fraction), cos(theta_fraction), 0, 0],
            [0, 0, 1, h_fraction],
            [0, 0, 0, 1]
        ])

        self.joint_angles = np.zeros(len(self.Slist[0]))
        self.joint_angles[1] = np.pi/2
        self.joint_angles[3] = np.pi/2
        self.joint_angles[5] = np.pi/2

        start_forward_kinematics = FKinBody(self.M, self.Blist, self.joint_angles)

        start_position = start_forward_kinematics[:4, 3]
        start_orientation = Rotation.from_matrix(start_forward_kinematics[:3, :3]).as_quat()

        end_forward_kinematics = np.matmul(intermediate_rotation, start_forward_kinematics)

        end_position = end_forward_kinematics[:4, 3]
        end_orientation = Rotation.from_matrix(end_forward_kinematics[:3, :3]).as_quat()

        self.publish_pose_simple_robot(end_position, end_orientation)
        if self.index == 1:
            self.publish_pose_target_pose(end_position, end_orientation)

        self.construct_path(end_position)

    def construct_path(self, vector):
        # RViz
        pose = PoseStamped()
        pose.header.frame_id = "map"
        pose.header.stamp = self.get_clock().now().to_msg()
        pose.pose.position.x = vector[0]
        pose.pose.position.y = vector[1]
        pose.pose.position.z = vector[2]
        pose.pose.orientation.w = 1.0

        #If the path is completed clear it
        if self.phi == 0:
            self.path.poses.clear()
        else:
            self.path.poses.append(pose)

        # Publish the path
        self.path_publisher_.publish(self.path)

        # Call the function publishing the marker
        self.publish_marker(pose.pose)

    def publish_marker(self, current_pose):
        # RViz
        marker = Marker()
        marker.header.frame_id = "map"
        marker.header.stamp = self.get_clock().now().to_msg()
        marker.type = Marker.SPHERE
        marker.action = Marker.ADD
        marker.pose = current_pose
        marker.scale.x = 0.05
        marker.scale.y = 0.05
        marker.scale.z = 0.05
        marker.color.a = 1.0
        marker.color.r = 1.0
        marker.color.g = 0.0
        marker.color.b = 0.0

        # Publish the marker
        self.point_publisher_.publish(marker)

    def publish_angle(self, angle):
        angle_msg = Float64()
        angle_msg.data = angle * 180/ pi
        self.angle_publisher_.publish(angle_msg)
        # self.get_logger().info('Publishing: %f' %angle_radians)

    def publish_height(self, height):
        height_msg = Float64()
        height_msg.data = height
        self.height_publisher_.publish(height_msg)

    def publish_pose_simple_robot(self, vector, quaternion):
        pose_msg = PoseStamped()
        pose_msg.header.frame_id = "map"
        pose_msg.header.stamp = self.get_clock().now().to_msg()
        pose_msg.pose.position.x = vector[0]
        pose_msg.pose.position.y = vector[1]
        pose_msg.pose.position.z = vector[2]
        pose_msg.pose.orientation.x = quaternion[0]
        pose_msg.pose.orientation.y = quaternion[1]
        pose_msg.pose.orientation.z = quaternion[2]
        pose_msg.pose.orientation.w = quaternion[3]
        self.pose_publisher_simple_robot_.publish(pose_msg)

    def publish_pose_target_pose(self, vector, quaternion):    
        # Create a Pose message for the target pose
        pose_msg = Pose()
        pose_msg.position.x = vector[0]
        pose_msg.position.y = vector[1]
        pose_msg.position.z = vector[2]
        pose_msg.orientation.x = quaternion[0]
        pose_msg.orientation.y = quaternion[1]
        pose_msg.orientation.z = quaternion[2]
        pose_msg.orientation.w = quaternion[3]
        self.pose_publisher_kuka_robot_.publish(pose_msg)
        # if self.phi == 0.0:
        #     time.sleep(1)
        


def main(args = None):
    rclpy.init(args = args)
    node = Forward_Kinematics_Node_2Dof()
    rclpy.spin(node)
    node.destroy_node()
    rclpy.shutdown()

if __name__ == '__main__':
    main()